#include <iostream>
#include <cstdlib>
#include <cctype>
#include <csignal>

#include "zero.hh"
#include "kertoimet.hh"
#include "vakiot.hh"
#include "yksikot.hh"
#include "yksikko.hh"
#include "summa.hh"
#include "tulo.hh"

extern void Dump(elemp res);

Kertoimet         Kert;
Yksikot           Yks;
Yksikkomuunnokset Ymu(Yks);
Vakiot            Vak(Yks);

static bool VirallinenTyyli = false;
static bool Optimization    = true;
        
class BadInput
{
	string desc;
public:
	BadInput() { }
	BadInput(const char *s) : desc(s) { }
	const string& value() const { return desc; }
};
class Parser
{
public:
    static void delblank(string &s)
    {
        while(s.size() && isspace(s[0]))s.erase(0, 1);
    }
    static merknum fetchnum(string &str) /* const is not valid for static */
    {
        char *ep;
        const char *s = str.c_str();
        const char *const begin = s;
        
        real val = strtold(s, &ep);
        
        // cout.precision(35);
        // cout << "value read: " << val << endl;
        
        /* Got the value, next analyze the precision */
        unsigned precis = 0;
        
        const char *q = s;
        if(*q == '+' || *q == '-')++q;
        bool pilkku = false;
        unsigned nollia=0;
        for(; q < ep; ++q)
        {
            if(*q == '0') { if(pilkku&&precis)++precis;else ++nollia; continue; }
            if(*q >= '1' && *q <= '9')
            {
                if(!pilkku)precis += nollia;
                ++precis;
                nollia = 0;
                continue;
            }
            /* Do not analyze the exponent */
            if(*q == 'E' || *q == 'e')break;
            if(*q == '.')
            {
                if(precis)precis += nollia;
                pilkku = true;
                nollia = 0;
            }
        }
        if(VirallinenTyyli || pilkku)
        {
            // Virallisessa 1000:n tarkkuus on 4
            // Joka tapauksessa 0.100:n tarkkuus on 3
            precis += nollia;
        }
        
        str.erase(0, ep-begin);
        delblank(str);
        if(str.size() && str[0]=='')
        {
        	str.erase(0, 1);
        	if(val > 0)
        		val += 0.5;
        	else
        		val -= 0.5;
	        delblank(str);
        }
        if(str.size() && str[0]=='#')
        {
        	str.erase(0, 1);
        	precis = TARKKUUS;
	        delblank(str);
        }
        return merknum(val, precis);
    }
    static tulo *ParseTulo(string &s, merknum &kerroin)
    {
        tulo *tulotmp = new tulo;
        
        bool jako = false;
        
        for(;;)
        {
            delblank(s);
            if(!s.size())break;
            
            unsigned a, b;
            int k = 0;
            
            elemp res;
            
            merknum ker = fetchnum(s);
            if(!ker.valid())ker = 1;
            
            //cout << "ker[" << ker.value() << "," << ker.precision() << ']';
            
            a = Kert.Test(s);
            if(a && (b = Yks.Test(s, a)) && Yks.SI(Yks.Value(s, a))) // Ei SI-kertoimettomia.
            {
                k = Kert.Value(s);
                /* k:n "yksikk": 10^ */
            YksLabel:
                char y = Yks.Value(s, a);
                s.erase(0, a+b);
                
                ker *= pow10l(k);
                res = new yksikko(y);
            }
            else if((a = Vak.Test(s)))
            {
                //cout << "(max precision with " << s << ")";
                res = &*Vak.Value(s);
                s.erase(0, a);
                
                ker.maxprecision();
            }
            else /* a=0 */
            {
                b = Yks.Test(s);
                if(b)
                {
                    /* k = 0; */
                    goto YksLabel;
                }
                /* a=0, b=0 */
                res = NULL;
                
                if(s.size() && s[0] == '(')
                {
                    s.erase(0, 1);
                    res = ParseSumma(s);
                    delblank(s);
                    if(s[0] != ')')throw BadInput("Expected: ')'");
                    s.erase(0, 1);
                }
            }
            
            real pot = 1;
            delblank(s);
            if(s.size())
            {
            	if(s[0]=='^')
	            {
    	            s.erase(0, 1);
        	        delblank(s);
	                pot = fetchnum(s).value();
    	        }
    	    }
    	    while(s.size())
    	    {
				if(s[0]!='' && s[0]!='')break;
				if(pot==1)pot=0;
				pot=pot*10;
				pot += (s[0]=='') ? 2 : 3;
  	        	s.erase(0, 1);
	    	}
	    	
            if(res)
            {
                tulotmp->insert(res, jako ? -pot : pot);
                if(jako)
                	kerroin /= ker;
                else
                	kerroin *= ker;
            }
            else
            {
	            if(jako)
    	            kerroin /= powl(ker, pot);
        	    else
            	    kerroin *= powl(ker, pot);
            	kerroin.combine(ker);
            }
            
            jako = false;
            if(s.size() && s[0]=='*')
            {
                s.erase(0, 1);
            }
            else if(s.size() && s[0]=='/')
            {
                jako = true;
                s.erase(0, 1);
            }
            else if(!res || (s.size() && (s[0]=='-' || s[0]=='+')))break;
        }
        return tulotmp;
    }
    static elemp ParseSumma(string &s)
    {
        summa *summatmp = new summa;
        bool miinus = false;
        for(;;)
        {
            delblank(s);
            if(!s.size())break;
            
            merknum kerroin = 1;
            tulo *p = ParseTulo(s, kerroin);
            if(!p)break;
            
            if(p->size())
                summatmp->insert(p, miinus ? -kerroin : kerroin);
            else if(miinus)
                summatmp->lisays -= kerroin;
            else
                summatmp->lisays += kerroin;
            
            delblank(s);
            if(s.size() && s[0]=='+')
            {
                miinus = false;
                s.erase(0, 1);
            }
            else if(s.size() && s[0]=='-')
            {
                miinus = true;
                s.erase(0, 1);
            }
            else break;
        }
        return summatmp;
    }
    
    static void DumpMerkNum(merknum e)
    {
    	if(e.value() < 0)
    	{
    		cout << '-';
    		e.negate();
    	}
    	
    	const unsigned precision = e.precision();
    	
    	/*
		Ekaksi kerroin sovitetaan sellaiseksi, ett luku on >= 1, mutta <= pow10(tarkkuus).
		Kokonaislukuosan desimaalien mr on log10(luku) (pyristettyn ylspin).
		Tmn jlkeen selvitetn tarvittavien desimaalien mr, ja pyristetn.
		Lopuksi tulostetaan tarvittaessa *10^jotain, mikli kerrottiin jollain.
    	*/
    	
    	// Se on nyt sitten positiivinen.
    	real v = e.value();
    	int kerroin = 0;
    	if(v < 1)
    	{
    		/* Ei tykt 0.00000001 -luvuista. 0.001 on viel siedettv. */
    		if(VirallinenTyyli || v < 0.001)
    		{
	    		// Pyrist negatiivisen luvun alaspin
    			kerroin = (int)floorl(log10l(v));
    			//cout << endl << v << "/10^" << kerroin;
    			v /= pow10l(kerroin);
    			//cout << "=" << v << endl;
    		}
    	}
    	else
    	{
    		real rajaarvo = pow10l(precision);
    		const int kokotarkkuus = (int)floorl(log10l(v)) + 1;
    		const int desimtarkkuus = precision - kokotarkkuus;
    		if(v + 9*pow10l(desimtarkkuus-1) >= rajaarvo)
    		{
    			kerroin = (int)log10l(v / rajaarvo) + 1;
	    		v /= pow10l(kerroin);
    		}
    	}
    	
    	int desimadd = 0;
    	
    	if(!VirallinenTyyli)
    	{
	    	/*  Esim 500.14: tarkkuus = 2. Normaali tulos = 50E1;
	    	    v:n kokonaislukuosa olisi tss vaiheessa 50, desimaaliosa 0.14, kerroin 1 (10^1).
	    	*/
	    	// Hvitetn ylimriset desimaalit.
	    	{
	    		const int kokotarkkuus = (int)floorl(log10l(v)) + 1;
    			const int desimtarkkuus = precision - kokotarkkuus;
    			/*
    			cout << "\nBefore: " << v << " = "
    			     << kokotarkkuus << '.' << desimtarkkuus 
    			     << " of " << precision << endl;
    			*/
    			const real powi = pow10l(desimtarkkuus);
    			v *= powi;
    			v += 1E-13L;
    			modfl(v, &v);
    			v /= powi;
    			
    			//cout << "After: " << v << endl;
    			// v -= fmodl(v, pow10l(-desimtarkkuus));
	    	}
	    	// Seuraavaksi pit selvitt, montako nollaa tss on.
	    	unsigned nollamaara;
	    	// Selvitmme sen loopilla. TODO: Keksi parempi menetelm.
	    	for(nollamaara=0; nollamaara<precision; ++nollamaara)
	    		if(fmodl(v, pow10l(nollamaara+1)) != 0)
	    			break;
	    	//cout << "Siin siis on " << nollamaara << " nollaa, luvussa " << v << " siis.\n";
	    	unsigned merkitsevia = (unsigned)log10l(v) + 1 - nollamaara; // miinus nollien mr. 
	    	//cout << "Merkitsevi: " << merkitsevia << endl;
	    	/*
	    	 Nyt siis jos luku on 500, 3 - 2 = 1.
	    	 Jos merkitsevi on enemmn kuin niit pitisi olla...
	    	 Ok, se on mahdotonta. Luku on jaettu jo tarpeeksi pieneksi.
	    	 Jos merkitsevi on liian vhn, on ongelma ksill.
	    	 Tm on siis esimerkiksi, jos luku on 500, ja halutaan 3 merkitsev.
	    	 Tllin on pakko ilmoittaa 5.00E2.
	    	*/
	    	if(merkitsevia < precision)
	    	{
	    		unsigned ero = precision - merkitsevia;
	    		if(ero <= nollamaara)
	    		{
		    		kerroin += ero;
		    		v /= pow10l(ero);
	    			// cout << "ero = " << ero << ", uusi v on nyt " << v << endl;
	    		}
	    	}
	    	else if(merkitsevia == precision)
	    	{
	    		// Jos luku on 500, ja halutaan 1 merkitsev.
	    		// Virallinen tyyli tuotta tuosta 5E2.
	    		// Koska pilkkua edeltvt nollat eivt ole nyt merkitsevi,
	    		// mys "500" on sallittu muoto.
	    		if(kerroin > 0)
	    		{
	    			v *= pow10l(kerroin);
	    			desimadd = kerroin;
	    			kerroin = 0;
	    		}
	    	}	    	
	    }
	    const int kokotarkkuus = (int)floorl(log10l(v)) + 1;
    	const int desimtarkkuus = precision - kokotarkkuus + desimadd;
    	
        char Buf[256];
        sprintf(Buf, "%.*Lf", desimtarkkuus, v);
        cout << Buf;
        if(kerroin != 0)
        {
        	sprintf(Buf, "E%d", kerroin);
        	cout << Buf;
        }
        //cout << '(' << kokotarkkuus << " of " << precision << ", desimtarkkuus=" << desimtarkkuus << ')';
    }
    static int SuggestKerroin(const merknum &e, bool SentitYms = false)
    {
    	real tmp = e.value();
    	int ker = 0;
    	
    	// Tm on tarkempaa kuin 10-jakolooppi
   		ker = (int)floorl(log10l(tmp));
    	
    	if(!SentitYms || ker < -6 || ker > 0)
    		if(ker < 0)
    			ker = int((ker-2)/3)*3;
    		else
	    		ker = int(ker/3)*3;
    	
    	if(ker > 18)ker = 18;
    	else if(ker < -18)ker = -18;
    	
    	return ker;
    }
    
    static void DumpResult(elemp result, int suggkerr = 0)
    {
        if(result->type=='+' || result->type=='*')
            cout << "\33[1;32m" << result->type << "\33[m";
        cout << flush;
        switch(result->type)
        {
            case '+':
            {
                summa &s = *(summa *)&*result;
                
                bool iniplus = false;
                
                if(s.lisays != 0 || s.elems.size() == 0)
                {
                    DumpMerkNum(s.lisays);
                    iniplus = true;
                }
                
                for(summa::elemap::const_iterator i = s.elems.begin(); i != s.elems.end(); ++i)
                {
                	merknum tmp = i->second;
                	bool si = false;
                	bool cent = false;
                    switch(i->first->type)
                    {
                    	case 'g':
	                    	tmp /= 1000;
	                    	break;
	                    case '*':
	                    {
                            tulo &s = *(tulo *)&*(i->first);
                            tulo::elemap::const_iterator j;
                            bool eka = true;
                            for(j = s.elems.begin(); j != s.elems.end(); ++j)
                            {
                            	if(j->first->type=='g')
                            		tmp /= powl(1000, j->second);
                            	else if(j->first->type!='+' && j->first->type!='*' && eka)
                            	{
	                            	Yksikot::si_case s = Yks.SI(j->first->type);
	                            	if(s != Yksikot::no_si)
	                            	{
	                            		if(s == Yksikot::si_c)cent = true;
	                            		si = true;
	                            	}
                            	}
                            	eka = false;
                            }
	                    	break;
	                    }
	                    case '+':
	                    	break;
	                    default:
	                    {
							Yksikot::si_case s = Yks.SI(i->first->type);
	                        if(s != Yksikot::no_si)
	                        {
	                        	if(s == Yksikot::si_c)cent = true;
	                        	si = true;
	                        }
	                    }
	                }
	                
	                int suggkerr = 0;	                
	                if(si)
	                {
	                	suggkerr = SuggestKerroin(tmp, cent);
	                	tmp /= pow10l(suggkerr);
	                }
                    if(tmp >= 0)
                    {
                        if(iniplus)cout << " + ";
                        DumpMerkNum(tmp);
                    }
                    else
                    {
                        cout << " - ";
                        DumpMerkNum(-tmp);
                    }
                    cout << ' ';
                    DumpResult(i->first, suggkerr);
                    iniplus = true;
                }
                break;
            }
            case '*':
            {
                tulo &s = *(tulo *)&*result;
                
                tulo::elemap::const_iterator i;
                
                bool eka=true;
                
                for(i = s.elems.begin(); i != s.elems.end(); ++i)
                {
                    if(i->second < 0)continue;
                    bool plus = i->first->type=='+' /*&& ((summa *)&*i->first)->size()*/;
                    if(plus)cout << '(';
                    if(suggkerr && Yks.SI(i->first->type) != Yksikot::no_si)
                    {
                    	cout << Kert.Find(suggkerr);
                    	//cout << '(' << suggkerr << ')';
                    	suggkerr = 0;
                    }
                    DumpResult(i->first);
                    if(plus)cout << ')';
                    if(i->second == 2)cout << '';
                    else if(i->second == 3)cout << '';
                    else if(i->second != 1)cout << '^' << i->second;
                    eka = false;
                }
                for(i = s.elems.begin(); i != s.elems.end(); ++i)
                {
                    if(i->second >= 0)continue;
                    if(eka) { cout << '1'; eka = false; }
                    cout << '/';
                    bool plus = i->first->type=='+' /*&& ((summa *)&*i->first)->size()*/;
                    if(plus)cout << '(';
                    if(suggkerr && Yks.SI(i->first->type) != Yksikot::no_si)
                    {
                    	cout << Kert.Find(suggkerr);
                    	//cout << '(' << suggkerr << ')';
                    	suggkerr = 0;
                    }
                    DumpResult(i->first);
                    if(plus)cout << ')';
                    if(i->second == -2)cout << '';
                    else if(i->second == -3)cout << '';
                    else if(i->second != -1)cout << '^' << i->second;
                }
                
                if(!s.size())cout << '1';
                
                break;
            }
            
            default:
            	if(suggkerr)
                {
                	cout << Kert.Find(suggkerr);
                	//cout << '(' << suggkerr << ')';
                    suggkerr = 0;
                }
            	if(result->type=='g')cout << 'k'; // Grammat aina kilogrammoiksi
            	//cout << result->type;
                cout << Yks[result->type];
        }
        cout << flush;
    }
    
    static void Parse(string s)
    {
        elemp result;
        bool vihre = false;
        try
        {
        	result = ParseSumma(s);
        }
        catch(BadInput &t)
        {
        	cout << "ce: " << t.value() << endl;
        	vihre = true;
        	result = NULL;
        }
        if(s.size())
        {
            cout << "ce: Parse error before \"" << s << "\"";
            if(!ispunct(s[0]))cout << " (unrecognized identifier)";
            cout << endl;
            vihre = true;
        }
        if(vihre)return;
        
        if(Optimization)result = result->Expand();
        
        cout << "Result: ";
        Dump(result);        
    }
};

static Parser p;
void Dump(elemp res)
{
    p.DumpResult(res);
    cout << endl;
}

static void SegTrap(int sig)
{
	cout << "Error: Unexpected input\n";
	raise(sig);
}

#define VERSIONSTR \
    "ce "VERSION" copyright (C) 1992,2000 Bisqwit (http://iki.fi/bisqwit/)\n" \
    "This program is under GPL. ce-"VERSION".{rar,zip,tar.{gz,bz2}}\n" \
    "are available at the homepage of the author.\n"

#include "args.hh"
class Handle : public arghandler
{
	bool Help;
    string opt_h(const string &s)
    {
    	Help = true;
    	return s;
    }
    string opt_V(const string &)
    {
    	cout << VERSIONSTR;
		exit(0);
    }
    #define bool_opt(yes, no, var) string yes(const string &s) { var = true; return s; } \
                                   string no(const string &s) { var = false; return s; }
    bool_opt(opt_m1, opt_m0, VirallinenTyyli)
    bool_opt(opt_s1, opt_s0, SummaDebug)
    bool_opt(opt_t1, opt_t0, TuloDebug)
    bool_opt(opt_c1, opt_c0, CombineDebug)
    bool_opt(opt_o,  opt_n,  Optimization)
    
    #undef bool_opt
public:
	string expression;
public:
	Handle(const char *defopts, int argc, const char *const *argv)
	: arghandler(defopts, argc, argv),
		Help(false)
	{
		add("-h",  "--help",        "This text", &(Handle::opt_h));
		add("-m0", "--no-stdsignif","500 contains one significant digit (default)", &(Handle::opt_m0));
		add("-m1", "--stdsignif",   "500 contains three significant digits", &(Handle::opt_m1));
		add("-n",  "--no-optimization", "Supresses the formula expansion", &(Handle::opt_n));
		add("-o",  "--optimization",  "Optimizes the formula (default)", &(Handle::opt_o));
		add("-s0", "--no-summadebug", "(default)", &(Handle::opt_s0));
		add("-s1", "--summadebug",   "Debygs sum operations", &(Handle::opt_s1));
		add("-t0", "--no-tulodebug", "(default)", &(Handle::opt_t0));
		add("-t1", "--tulodebug",   "Debygs product operations", &(Handle::opt_t1));
		add("-c0", "--no-combinedebug", "(default)", &(Handle::opt_c0));
		add("-c1", "--combinedebug",   "Debygs combining operations", &(Handle::opt_c1));
		add("-V",  "--version",     "Version info", &(Handle::opt_V));
		parse();
	}
	virtual void defarg(const string &s)
	{
		// cout << "defarg(\"" << s << "\")\n";
		if(s.size())
		{
			if(expression.size())
				expression += ' ';
			expression += s;
		}
	}
    virtual void parse()
    {
    	arghandler::parse();
		if(Help)
		{
            cout << "Usage: " << a0 << " [options] <expression>\n"
                    "\nOptions:\n";
            
            listoptions();
            
            cout << endl;
            
            exit(0);
		}
    }
};

int main(int argc, const char *const *argv)
{
    signal(SIGSEGV, SegTrap);
    
    Handle parameters("", argc, argv);
    string s = parameters.expression;
    
    if(!s.size())
    {
        cout << "Anna lauseke: " << flush;
        getline(cin, s);
    }
    
    // cout << "Parsing: '" << s << "'\n";
        
    p.Parse(s);
    
    return 0;
}
